MODULE step
   !!======================================================================
   !!                       ***  MODULE step  ***
   !! Time-stepping    : manager of the ocean, tracer and ice time stepping
   !!======================================================================
   !! History :  OPA  !  1991-03  (G. Madec)  Original code
   !!             -   !  1991-11  (G. Madec)
   !!             -   !  1992-06  (M. Imbard)  add a first output record
   !!             -   !  1996-04  (G. Madec)  introduction of dynspg
   !!             -   !  1996-04  (M.A. Foujols)  introduction of passive tracer
   !!            8.0  !  1997-06  (G. Madec)  new architecture of call
   !!            8.2  !  1997-06  (G. Madec, M. Imbard, G. Roullet)  free surface
   !!             -   !  1999-02  (G. Madec, N. Grima)  hpg implicit
   !!             -   !  2000-07  (J-M Molines, M. Imbard)  Open Bondary Conditions
   !!   NEMO     1.0  !  2002-06  (G. Madec)  free form, suppress macro-tasking
   !!             -   !  2004-08  (C. Talandier) New trends organization
   !!             -   !  2005-01  (C. Ethe) Add the KPP closure scheme
   !!             -   !  2005-11  (G. Madec)  Reorganisation of tra and dyn calls
   !!             -   !  2006-01  (L. Debreu, C. Mazauric)  Agrif implementation
   !!             -   !  2006-07  (S. Masson)  restart using iom
   !!            3.2  !  2009-02  (G. Madec, R. Benshila)  reintroduicing z*-coordinate
   !!             -   !  2009-06  (S. Masson, G. Madec)  TKE restart compatible with key_cpl
   !!            3.3  !  2010-05  (K. Mogensen, A. Weaver, M. Martin, D. Lea) Assimilation interface
   !!             -   !  2010-10  (C. Ethe, G. Madec) reorganisation of initialisation phase + merge TRC-TRA
   !!                 !  2011-05  (S. Pickles) dCSE NEMO optimisations - z index first
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   stp             : OPA system time-stepping
   !!----------------------------------------------------------------------
   USE step_oce         ! time stepping definition modules 
#if defined key_top
   USE trcstp           ! passive tracer time-stepping      (trc_stp routine)
#endif
#if defined key_agrif
   USE agrif_opa_sponge ! Momemtum and tracers sponges
#endif
   USE asminc           ! assimilation increments    (tra_asm_inc, dyn_asm_inc routines)
   USE timing, ONLY: timing_start, timing_stop, timing_reset, timing_disable
   USE arpdebugging, ONLY: dump_array
   IMPLICIT NONE
   PRIVATE

   PUBLIC   stp   ! called by opa.F90

   !! * Control permutation of array indices
   !! DCSE_NEMO: warning! dom_oce and zdf_oce public variables are made available
   !! through the use of step_oce
#  include "dom_oce_ftrans.h90"
#  include "zdf_oce_ftrans.h90"

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "zdfddm_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OPA 3.3 , NEMO Consortium (2010)
   !! $Id$
   !! Software governed by the CeCILL licence     (NEMOGCM/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------
CONTAINS

#if defined key_agrif
   SUBROUTINE stp( )
      INTEGER             ::   kstp   ! ocean time-step index
#else
   SUBROUTINE stp( kstp )
      INTEGER, INTENT(in) ::   kstp   ! ocean time-step index
#endif
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE stp  ***
      !!                      
      !! ** Purpose : - Time stepping of OPA (momentum and active tracer eqs.)
      !!              - Time stepping of LIM (dynamic and thermodynamic eqs.)
      !!              - Tme stepping  of TRC (passive tracer eqs.)
      !! 
      !! ** Method  : -1- Update forcings and data  
      !!              -2- Update ocean physics 
      !!              -3- Compute the t and s trends 
      !!              -4- Update t and s 
      !!              -5- Compute the momentum trends
      !!              -6- Update the horizontal velocity
      !!              -7- Compute the diagnostics variables (rd,N2, div,cur,w)
      !!              -8- Outputs and diagnostics
      !!----------------------------------------------------------------------
      INTEGER ::   jk       ! dummy loop indice
      INTEGER ::   indic    ! error indicator if < 0
#if defined key_z_first
      INTEGER ::   ji, jj   ! dummy loop indices
#endif
      INTEGER, PARAMETER :: DUMP_LEVEL = 26 ! ARPDBG - level to dump to disk
      !! ---------------------------------------------------------------------
                         CALL timing_start('Step') 

!                      IF(kstp==1)CALL dump_array(kstp,'tmask_i_stp',tmask_i, &
!                                                 withHalos=.TRUE.)

#if defined key_agrif
      kstp = nit000 + Agrif_Nb_Step()
!      IF ( Agrif_Root() .and. lwp) Write(*,*) '---'
!      IF (lwp) Write(*,*) 'Grid Number',Agrif_Fixed(),' time step ',kstp
# if defined key_iomput
      IF( Agrif_Nbstepint() == 0 )   CALL iom_swap
# endif   
#endif   
                             indic = 0                ! reset to no error condition
      IF( kstp /= nit000 )   CALL day( kstp )         ! Calendar (day was already called at nit000 in day_init)
                             CALL iom_setkt( kstp )   ! say to iom that we are at time step kstp

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Update data, open boundaries, surface boundary condition (including sea-ice)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL timing_start('Boundaries') 
      IF( lk_dtatem  )   CALL dta_tem( kstp )         ! update 3D temperature data
      IF( lk_dtasal  )   CALL dta_sal( kstp )         ! update 3D salinity data
                         CALL sbc    ( kstp )         ! Sea Boundary Condition (including sea-ice)
      IF( lk_obc     )   CALL obc_dta( kstp )         ! update dynamic and tracer data at open boundaries
      IF( lk_obc     )   CALL obc_rad( kstp )         ! compute phase velocities at open boundaries
      IF( lk_bdy     )   CALL bdy_dta_frs( kstp )     ! update dynamic and tracer data for FRS conditions (BDY)
                         ! ARP - no 'section' arg here so the time taken
                         ! in this region (which can involve IO) is 
                         ! subtracted from the time
                         ! taken to do the whole step.
                         CALL timing_stop('Boundaries') 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      !  Ocean dynamics : ssh, wn, hdiv, rot                                 !
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL timing_start('Ocean Dyn.') 
                         CALL ssh_wzv( kstp )         ! after ssh & vertical velocity
                         CALL timing_stop('Ocean Dyn.','section') 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Ocean physics update                (ua, va, ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL timing_start('Ocean Phys.') 
                         CALL bn2( tsb, rn2b )        ! before Brunt-Vaisala frequency
                         CALL bn2( tsn, rn2  )        ! now    Brunt-Vaisala frequency
                         CALL timing_stop('Ocean Phys.','section') 
      !
      !  VERTICAL PHYSICS  
                         CALL timing_start('Vert. Phys.') 
                         CALL zdf_bfr( kstp )         ! bottom friction
                         
      !                                               ! Vertical eddy viscosity and diffusivity coefficients
      IF( lk_zdfric  )   CALL zdf_ric( kstp )            ! Richardson number dependent Kz
      IF( lk_zdftke  )   CALL zdf_tke( kstp )            ! TKE closure scheme for Kz
      IF( lk_zdfgls  )   CALL zdf_gls( kstp )            ! GLS closure scheme for Kz
      IF( lk_zdfkpp  )   CALL zdf_kpp( kstp )            ! KPP closure scheme for Kz
      IF( lk_zdfcst  )   THEN                            ! Constant Kz (reset avt, avm[uv] to the background value)
         avt (:,:,:) = rn_avt0 * tmask(:,:,:)
         avmu(:,:,:) = rn_avm0 * umask(:,:,:)
         avmv(:,:,:) = rn_avm0 * vmask(:,:,:)
      ENDIF
      IF( ln_rnf_mouth ) THEN                         ! increase diffusivity at rivers mouths
#if defined key_z_first
         DO ji = 1, jpi
            DO jj = 1, jpj
               DO jk = 2, nkrnf 
                  avt(ji,jj,jk) = avt(ji,jj,jk) + 2.e0 * rn_avt_rnf * rnfmsk(ji,jj) * tmask(ji,jj,jk)
               END DO
            END DO
         END DO
#else
         DO jk = 2, nkrnf
            avt(:,:,jk) = avt(:,:,jk) + 2.e0 * rn_avt_rnf * rnfmsk(:,:) * tmask(:,:,jk)
         END DO
#endif
      ENDIF
      IF( ln_zdfevd  )   CALL zdf_evd( kstp )         ! enhanced vertical eddy diffusivity

      IF( lk_zdftmx  )   CALL zdf_tmx( kstp )         ! tidal vertical mixing

      IF( lk_zdfddm .AND. .NOT. lk_zdfkpp )   &
         &               CALL zdf_ddm( kstp )         ! double diffusive mixing
         
                         CALL zdf_mxl( kstp )         ! mixed layer depth

                                                      ! write TKE or GLS information in the restart file
      IF( lrst_oce .AND. lk_zdftke )   CALL tke_rst( kstp, 'WRITE' )
      IF( lrst_oce .AND. lk_zdfgls )   CALL gls_rst( kstp, 'WRITE' )

                         CALL timing_stop('Vert. Phys.','section') 
      !
      !  LATERAL  PHYSICS 
      !
                         CALL timing_start('Lateral Phys.') 
      !
      IF( lk_ldfslp ) THEN                            ! slope of lateral mixing
                         CALL eos( tsb, rhd )                ! before in situ density
         IF( ln_zps )    CALL zps_hde( kstp, jpts, tsb, gtsu, gtsv,  &    ! Partial steps: before horizontal gradient
            &                                      rhd, gru , grv  )      ! of t, s, rd at the last ocean level
         IF( ln_traldf_grif ) THEN                           ! before slope for Griffies operator
                         CALL ldf_slp_grif( kstp )
         ELSE
                         CALL ldf_slp( kstp, rhd, rn2b )     ! before slope for Madec operator
         ENDIF
      ENDIF
#if defined key_traldf_c2d
      IF( lk_traldf_eiv )   CALL ldf_eiv( kstp )      ! eddy induced velocity coefficient
#endif
                         CALL timing_stop('Lateral Phys.','section') 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! diagnostics and outputs             (ua, va, ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL timing_start('Diagnostics') 
      IF( lk_floats  )   CALL flo_stp( kstp )         ! drifting Floats
      IF( lk_diahth  )   CALL dia_hth( kstp )         ! Thermocline depth (20 degres isotherm depth)
      IF( lk_diafwb  )   CALL dia_fwb( kstp )         ! Fresh water budget diagnostics
      IF( ln_diaptr  )   CALL dia_ptr( kstp )         ! Poleward TRansports diagnostics
      IF( lk_diaar5  )   CALL dia_ar5( kstp )         ! ar5 diag
                         CALL dia_wri( kstp )         ! ocean model: outputs
                         CALL timing_stop('Diagnostics') 

#if defined key_top
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Passive Tracer Model
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL trc_stp( kstp )         ! time-stepping
#endif

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Active tracers                              (ua, va used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                             CALL timing_start('Active tracers') 

                             tsa(:,:,:,:) = 0.e0            ! set tracer trends to zero

      IF(  ln_asmiau .AND. &
         & ln_trainc     )   CALL tra_asm_inc( kstp )       ! apply tracer assimilation increment
                             CALL tra_sbc    ( kstp )       ! surface boundary condition
      IF( ln_traqsr      )   CALL tra_qsr    ( kstp )       ! penetrative solar radiation qsr
      IF( ln_trabbc      )   CALL tra_bbc    ( kstp )       ! bottom heat flux
      IF( lk_trabbl      )   CALL tra_bbl    ( kstp )       ! advective (and/or diffusive) bottom boundary layer scheme
      IF( lk_tradmp      )   CALL tra_dmp    ( kstp )       ! internal damping trends
                             CALL tra_adv    ( kstp )       ! horizontal & vertical advection
      IF( lk_zdfkpp      )   CALL tra_kpp    ( kstp )       ! KPP non-local tracer fluxes
                             CALL tra_ldf    ( kstp )       ! lateral mixing
#if defined key_agrif
                             CALL tra_unswap
      IF(.NOT. Agrif_Root()) CALL Agrif_Sponge_tra          ! tracers sponge
                             CALL tra_swap
#endif
                             CALL tra_zdf    ( kstp )       ! vertical mixing and after tracer fields

      IF( ln_dynhpg_imp  ) THEN                             ! semi-implicit hpg (time stepping then eos)
         IF( ln_zdfnpc   )   CALL tra_npc( kstp )                ! update after fields by non-penetrative convection
                             CALL tra_nxt( kstp )                ! tracer fields at next time step
                             CALL eos    ( tsa, rhd, rhop )      ! Time-filtered in situ density for hpg computation
         IF( ln_zps      )   CALL zps_hde( kstp, jpts, tsa, gtsu, gtsv,  &    ! zps: time filtered hor. derivative
            &                                          rhd, gru , grv  )      ! of t, s, rd at the last ocean level
         
      ELSE                                                  ! centered hpg  (eos then time stepping)
                             CALL eos    ( tsn, rhd, rhop )      ! now in situ density for hpg computation
         IF( ln_zps      )   CALL zps_hde( kstp, jpts, tsn, gtsu, gtsv,  &    ! zps: now hor. derivative
            &                                          rhd, gru , grv  )      ! of t, s, rd at the last ocean level
         IF( ln_zdfnpc   )   CALL tra_npc( kstp )                ! update after fields by non-penetrative convection
                             CALL tra_nxt( kstp )                ! tracer fields at next time step
      ENDIF 
                             CALL tra_unswap                ! udate T & S 3D arrays  (to be suppressed)
                             CALL timing_stop('Active tracers','section') 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Dynamics                                    (ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                               CALL timing_start('Dynamics') 

                               ua(:,:,:) = 0.e0             ! set dynamics trends to zero
                               va(:,:,:) = 0.e0

      IF(  ln_asmiau .AND. &
         & ln_dyninc       )   CALL dyn_asm_inc( kstp )     ! apply dynamics assimilation increment

                               !CALL timing_start('dyn_adv') 
                               CALL dyn_adv( kstp )         ! advection (vector or flux form)
                               !CALL timing_stop('dyn_adv','section') 

!                               CALL dump_array(kstp,'ua_pre_vor',ua(:,:,DUMP_LEVEL), &
!                                               withHalos=.TRUE.,atStep=1)
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_adv SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

                               !CALL timing_start('dyn_vor') 
                               CALL dyn_vor( kstp )         ! vorticity term including Coriolis
                               !CALL timing_stop('dyn_vor','section') 

!                               CALL dump_array(kstp,'ua_pre_ldf',ua(:,:,DUMP_LEVEL), &
!                                               withHalos=.TRUE.,atStep=1)
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_vor SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

                               !CALL timing_start('dyn_ldf') 
                               CALL dyn_ldf( kstp )         ! lateral mixing
                               !CALL timing_stop('dyn_ldf','section') 
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_ldf SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

#if defined key_agrif
      IF(.NOT. Agrif_Root())   CALL Agrif_Sponge_dyn        ! momemtum sponge
#endif
                               !CALL dump_array(kstp,'ua_pre_hpg',ua(:,:,DUMP_LEVEL), &
                               !                withHalos=.TRUE.)

                               !CALL timing_start('dyn_hpg') 
                               CALL dyn_hpg( kstp )         ! horizontal gradient of Hydrostatic pressure
                               !CALL timing_stop('dyn_hpg','section') 

                               !CALL dump_array(kstp,'ua_pre_bfr',ua(:,:,26), &
                               !                withHalos=.TRUE.)
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_hpg SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

                               !CALL timing_start('dyn_bfr') 
                               CALL dyn_bfr( kstp )         ! bottom friction   
                               !CALL timing_stop('dyn_bfr','section') 
                               
!                               CALL dump_array(kstp,'ua_pre_zdf',ua(:,:,DUMP_LEVEL), &
!                                               withHalos=.TRUE.,atStep=1)
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_bfr SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

                               !CALL timing_start('dyn_zdf') 
                               CALL dyn_zdf( kstp )         ! vertical diffusion
                               !CALL timing_stop('dyn_zdf','section') 
                               
!                               CALL dump_array(kstp,'ua_pre_spg',ua(:,:,DUMP_LEVEL), &
!                                               withHalos=.TRUE.,atStep=1)
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_zdf SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

                               !CALL timing_start('dyn_spg') 
                               CALL dyn_spg( kstp, indic )  ! surface pressure gradient
                               !CALL timing_stop('dyn_spg','section') 
!                               CALL dump_array(kstp,'ua_spg',ua(:,:,DUMP_LEVEL), &
!                                               withHalos=.TRUE.,atStep=1)
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_spg SUM(ua)= ',SUM(ua),' at step=',kstp
#endif
                               !CALL timing_start('dyn_nxt') 
                               CALL dyn_nxt( kstp )         ! lateral velocity at next time step
                               !CALL timing_stop('dyn_nxt','section') 

                               !CALL timing_stop('dyn_nxt','section') 
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after dyn_nxt SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

                               !CALL timing_start('ssh_nxt') 
                               CALL ssh_nxt( kstp )         ! sea surface height at next time step
                               !CALL timing_stop('ssh_nxt','section') 
#if defined ARPDBGSUM
                               WRITE (*,*) narea,': ARPDBG: after ssh_nxt SUM(ua)= ',SUM(ua),' at step=',kstp
#endif

      IF( ln_diahsb        )   CALL dia_hsb( kstp )         ! - ML - global conservation diagnostics
      IF( lk_diaobs  )         CALL dia_obs( kstp )         ! obs-minus-model (assimilation) diagnostics
                               !                            ! (call after dynamics update)
                               CALL timing_stop('Dynamics','section') 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Control and restarts
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                               CALL timing_start('Control') 

                      !IF(kstp==1)CALL dump_array(kstp,'tmask_i_ctl',tmask_i, &
                      !                           withHalos=.TRUE.)

                               CALL stp_ctl( kstp, indic )
      IF( indic < 0        )   THEN
                               CALL ctl_stop( 'step: indic < 0' )
                               CALL dia_wri_state( 'output.abort', kstp )
      ENDIF
      IF( kstp == nit000   )   CALL iom_close( numror )     ! close input  ocean restart file
      IF( lrst_oce         )   CALL rst_write    ( kstp )   ! write output ocean restart file
      IF( lk_obc           )   CALL obc_rst_write( kstp )   ! write open boundary restart file
                               ! ARP - no 'section' arg here so the time taken
                               ! in this region (which can involve IO) is 
                               ! subtracted from the time
                               ! taken to do the whole step.
                               CALL timing_stop('Control') 
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Trends                              (ua, va, ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF( nstop == 0 ) THEN                         
         IF( lk_trddyn     )   CALL trd_dwr( kstp )         ! trends: dynamics 
         IF( lk_trdtra     )   CALL trd_twr( kstp )         ! trends: active tracers
         IF( lk_trdmld     )   CALL trd_mld( kstp )         ! trends: Mixed-layer 
         IF( lk_trdvor     )   CALL trd_vor( kstp )         ! trends: vorticity budget
      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Coupled mode
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF( lk_cpl           )   CALL sbc_cpl_snd( kstp )     ! coupled mode : field exchanges
      !
                               CALL timing_stop('Step', 'section') 

      !
      IF( kstp == nit000   )   CALL timing_reset            ! Exclude first step from timing
      IF( kstp == (nitend-2) ) CALL timing_disable()        ! Turn off timing two steps from end to avoid the IO
      !

   END SUBROUTINE stp

   !!======================================================================
END MODULE step
